@@ -6,7 +6,7 @@ from setuptools import Command
 
 from pex.bin.pex import build_pex, configure_clp, make_relative_to_root
 from pex.common import die
-from pex.compatibility import ConfigParser, StringIO, string
+from pex.compatibility import ConfigParser, StringIO, string, to_unicode
 from pex.variables import ENV
 
 
@@ -49,7 +49,7 @@ class bdist_pex(Command):  # noqa
 
     if isinstance(raw_entry_points, string):
       parser = ConfigParser()
-      parser.readfp(StringIO(raw_entry_points))
+      parser.readfp(StringIO(to_unicode(raw_entry_points)))
       if parser.has_section('console_scripts'):
         return dict(parser.items('console_scripts'))
     elif isinstance(raw_entry_points, dict):
