@@ -25,7 +25,7 @@ def die(msg, exit_code=1):
 def safe_copy(source, dest, overwrite=False):
   def do_copy():
     temp_dest = dest + uuid4().hex
-    shutil.copyfile(source, temp_dest)
+    shutil.copy(source, temp_dest)
     os.rename(temp_dest, dest)
 
   # If the platform supports hard-linking, use that and fall back to copying.
@@ -78,10 +78,27 @@ class MktempTeardownRegistry(object):
 _MKDTEMP_SINGLETON = MktempTeardownRegistry()
 
 
+class PermPreservingZipFile(zipfile.ZipFile, object):
+  """A ZipFile that works around https://bugs.python.org/issue15795"""
+
+  def _extract_member(self, member, targetpath, pwd):
+    result = super(PermPreservingZipFile, self)._extract_member(member, targetpath, pwd)
+    info = member if isinstance(member, zipfile.ZipInfo) else self.getinfo(member)
+    self._chmod(info, result)
+    return result
+
+  def _chmod(self, info, path):
+    # This magic works to extract perm bits from the 32 bit external file attributes field for
+    # unix-created zip files, for the layout, see:
+    #   https://www.forensicswiki.org/wiki/ZIP#External_file_attributes
+    attr = info.external_attr >> 16
+    os.chmod(path, attr)
+
+
 @contextlib.contextmanager
 def open_zip(path, *args, **kwargs):
   """A contextmanager for zip files.  Passes through positional and kwargs to zipfile.ZipFile."""
-  with contextlib.closing(zipfile.ZipFile(path, *args, **kwargs)) as zip:
+  with contextlib.closing(PermPreservingZipFile(path, *args, **kwargs)) as zip:
     yield zip
 
 
@@ -103,8 +120,7 @@ def safe_mkdir(directory, clean=False):
   """Safely create a directory.
 
   Ensures a directory is present.  If it's not there, it is created.  If it
-  is, it's a no-op.  no-op.  If clean is True, ensures the directory is
-  empty.
+  is, it's a no-op. If clean is True, ensures the directory is empty.
   """
   if clean:
     safe_rmtree(directory)
@@ -262,7 +278,7 @@ class Chroot(object):
   def copy(self, src, dst, label=None):
     """Copy file ``src`` to ``chroot/dst`` with optional label.
 
-    May raise anything shutil.copyfile can raise, e.g.
+    May raise anything shutil.copy can raise, e.g.
       IOError(Errno 21 'EISDIR')
 
     May raise ChrootTaggingException if dst is already in a fileset
@@ -271,7 +287,7 @@ class Chroot(object):
     dst = self._normalize(dst)
     self._tag(dst, label)
     self._ensure_parent(dst)
-    shutil.copyfile(src, os.path.join(self.chroot, dst))
+    shutil.copy(src, os.path.join(self.chroot, dst))
 
   def link(self, src, dst, label=None):
     """Hard link file from ``src`` to ``chroot/dst`` with optional label.
@@ -331,7 +347,7 @@ class Chroot(object):
   def delete(self):
     shutil.rmtree(self.chroot)
 
-  def zip(self, filename, mode='wb'):
-    with contextlib.closing(zipfile.ZipFile(filename, mode)) as zf:
+  def zip(self, filename, mode='w'):
+    with open_zip(filename, mode) as zf:
       for f in sorted(self.files()):
         zf.write(os.path.join(self.chroot, f), arcname=f, compress_type=zipfile.ZIP_DEFLATED)
