@@ -17,11 +17,12 @@ from pkg_resources import (
     find_distributions
 )
 
-from .common import die, open_zip, safe_mkdir, safe_rmtree
+from .common import die, open_zip, rename_if_empty, safe_mkdir, safe_rmtree
 from .interpreter import PythonInterpreter
 from .package import distribution_compatible
 from .pex_builder import PEXBuilder
 from .pex_info import PexInfo
+from .platforms import Platform
 from .tracer import TRACER
 from .util import CacheHelper, DistributionHelper
 
@@ -48,7 +49,7 @@ class PEXEnvironment(Environment):
           safe_rmtree(explode_tmp)
           raise
       TRACER.log('Renaming %s to %s' % (explode_tmp, explode_dir))
-      os.rename(explode_tmp, explode_dir)
+      rename_if_empty(explode_tmp, explode_dir)
     return explode_dir
 
   @classmethod
@@ -57,10 +58,21 @@ class PEXEnvironment(Environment):
     TRACER.log('Adding to the head of sys.path: %s' % new_code_path)
     sys.path.insert(0, new_code_path)
     for name, module in sys.modules.items():
-      if hasattr(module, "__path__"):
+      if hasattr(module, '__path__'):
         module_dir = os.path.join(new_code_path, *name.split("."))
         TRACER.log('Adding to the head of %s.__path__: %s' % (module.__name__, module_dir))
-        module.__path__.insert(0, module_dir)
+        try:
+          module.__path__.insert(0, module_dir)
+        except AttributeError:
+          # TODO: This is a temporary bandaid for an unhandled AttributeError which results
+          # in a startup crash. See https://github.com/pantsbuild/pex/issues/598 for more info.
+          TRACER.log(
+            'Failed to insert %s: %s.__path__ of type %s does not support insertion!' % (
+              module_dir,
+              module.__name__,
+              type(module.__path__)
+            )
+          )
 
   @classmethod
   def write_zipped_internal_cache(cls, pex, pex_info):
@@ -82,17 +94,20 @@ class PEXEnvironment(Environment):
           distribution_name, dist_digest))
         if os.path.exists(cached_location):
           dist = DistributionHelper.distribution_from_path(cached_location)
-          existing_cached_distributions.append(dist)
-          continue
+          if dist is not None:
+            existing_cached_distributions.append(dist)
+            continue
         else:
           dist = DistributionHelper.distribution_from_path(os.path.join(pex, internal_dist_path))
-          if DistributionHelper.zipsafe(dist) and not pex_info.always_write_cache:
-            zip_safe_distributions.append(dist)
-            continue
+          if dist is not None:
+            if DistributionHelper.zipsafe(dist) and not pex_info.always_write_cache:
+              zip_safe_distributions.append(dist)
+              continue
 
         with TRACER.timed('Caching %s' % dist):
           newly_cached_distributions.append(
             CacheHelper.cache_distribution(zf, internal_dist_path, cached_location))
+
     return existing_cached_distributions, newly_cached_distributions, zip_safe_distributions
 
   @classmethod
@@ -114,8 +129,24 @@ class PEXEnvironment(Environment):
     self._activated = False
     self._working_set = None
     self._interpreter = interpreter or PythonInterpreter.get()
+    self._inherit_path = pex_info.inherit_path
+    self._supported_tags = []
+
+    platform = Platform.current()
+    platform_name = platform.platform
     super(PEXEnvironment, self).__init__(
-        search_path=sys.path if pex_info.inherit_path else [], **kw)
+      search_path=[] if pex_info.inherit_path == 'false' else sys.path,
+      # NB: Our pkg_resources.Environment base-class wants the platform name string and not the
+      # pex.platform.Platform object.
+      platform=platform_name,
+      **kw
+    )
+    self._target_interpreter_env = self._interpreter.identity.pkg_resources_env(platform_name)
+    self._supported_tags.extend(platform.supported_tags(self._interpreter))
+    TRACER.log(
+      'E: tags for %r x %r -> %s' % (self.platform, self._interpreter, self._supported_tags),
+      V=9
+    )
 
   def update_candidate_distributions(self, distribution_iter):
     for dist in distribution_iter:
@@ -124,7 +155,7 @@ class PEXEnvironment(Environment):
           self.add(dist)
 
   def can_add(self, dist):
-    return distribution_compatible(dist, self._interpreter, self.platform)
+    return distribution_compatible(dist, self._supported_tags)
 
   def activate(self):
     if not self._activated:
@@ -139,9 +170,15 @@ class PEXEnvironment(Environment):
     unresolved_reqs = set()
     resolveds = set()
 
+    environment = self._target_interpreter_env.copy()
+    environment['extra'] = list(set(itertools.chain(*(req.extras for req in reqs))))
+
     # Resolve them one at a time so that we can figure out which ones we need to elide should
     # there be an interpreter incompatibility.
     for req in reqs:
+      if req.marker and not req.marker.evaluate(environment=environment):
+        TRACER.log('Skipping activation of `%s` due to environment marker de-selection' % req)
+        continue
       with TRACER.timed('Resolving %s' % req, V=2):
         try:
           resolveds.update(working_set.resolve([req], env=self))
@@ -167,8 +204,12 @@ class PEXEnvironment(Environment):
         for dist in self._pex_info.distributions:
           TRACER.log('  - %s' % dist)
       if not self._pex_info.ignore_errors:
-        die('Failed to execute PEX file, missing compatible dependencies for:\n%s' % (
-            '\n'.join(map(str, unresolved_reqs))))
+        die(
+          'Failed to execute PEX file, missing %s compatible dependencies for:\n%s' % (
+            Platform.current(),
+            '\n'.join(str(r) for r in unresolved_reqs)
+          )
+        )
 
     return resolveds
 
@@ -189,6 +230,18 @@ class PEXEnvironment(Environment):
 
         if os.path.isdir(dist.location):
           with TRACER.timed('Adding sitedir', V=2):
+            if dist.location not in sys.path and self._inherit_path == "fallback":
+              # Prepend location to sys.path.
+              # This ensures that bundled versions of libraries will be used before system-installed
+              # versions, in case something is installed in both, helping to favor hermeticity in
+              # the case of non-hermetic PEX files (i.e. those with inherit_path=True).
+              #
+              # If the path is not already in sys.path, site.addsitedir will append (not prepend)
+              # the path to sys.path. But if the path is already in sys.path, site.addsitedir will
+              # leave sys.path unmodified, but will do everything else it would do. This is not part
+              # of its advertised contract (which is very vague), but has been verified to be the
+              # case by inspecting its source for both cpython 2.7 and cpython 3.7.
+              sys.path.insert(0, dist.location)
             site.addsitedir(dist.location)
 
         dist.activate()
