@@ -16,15 +16,16 @@ To use:
 
 import os
 import pkgutil
+import re
 import sys
 import zipimport
 
 import pkg_resources
 
 if sys.version_info >= (3, 3) and sys.implementation.name == "cpython":
-  import importlib._bootstrap as importlib_bootstrap
+  import importlib.machinery as importlib_machinery
 else:
-  importlib_bootstrap = None
+  importlib_machinery = None
 
 
 class ChainedFinder(object):
@@ -99,10 +100,39 @@ def _remove_finder(importer, finder):
 class WheelMetadata(pkg_resources.EggMetadata):
   """Metadata provider for zipped wheels."""
 
+  @classmethod
+  def _escape(cls, filename_component):
+    # See: https://www.python.org/dev/peps/pep-0427/#escaping-and-unicode
+    return re.sub("[^\w\d.]+", "_", filename_component, re.UNICODE)
+
   @classmethod
   def _split_wheelname(cls, wheelname):
-    split_wheelname = wheelname.split('-')
-    return '-'.join(split_wheelname[:-3])
+    # See: https://www.python.org/dev/peps/pep-0427/#file-name-convention
+    assert wheelname.endswith('.whl'), 'invalid wheel name: %s' % wheelname
+    split_wheelname = wheelname.rsplit('-', 5)
+    assert len(split_wheelname) in (5, 6), 'invalid wheel name: %s' % wheelname
+    distribution, version = split_wheelname[:2]
+    return '%s-%s' % (distribution, version)
+
+  @classmethod
+  def data_dir(cls, wheel_path):
+    """Returns the internal path of the data dir for the given wheel.
+
+    As defined https://www.python.org/dev/peps/pep-0427/#the-data-directory
+
+    :rtype: str
+    """
+    return '%s.data' % cls._split_wheelname(os.path.basename(wheel_path))
+
+  @classmethod
+  def dist_info_dir(cls, wheel_path):
+    """Returns the internal path of the dist-info dir for the given wheel.
+
+    As defined here: https://www.python.org/dev/peps/pep-0427/#the-dist-info-directory
+
+    :rtype: str
+    """
+    return '%s.dist-info' % cls._split_wheelname(os.path.basename(wheel_path))
 
   def _setup_prefix(self):
     path = self.module_path
@@ -112,7 +142,7 @@ class WheelMetadata(pkg_resources.EggMetadata):
         self.egg_name = os.path.basename(path)
         # TODO(wickman) Test the regression where we have both upper and lower cased package
         # names.
-        self.egg_info = os.path.join(path, '%s.dist-info' % self._split_wheelname(self.egg_name))
+        self.egg_info = os.path.join(path, self.dist_info_dir(self.egg_name))
         self.egg_root = path
         break
       old = path
@@ -214,8 +244,8 @@ def register_finders():
   # append the wheel finder
   _add_finder(pkgutil.ImpImporter, find_wheels_on_path)
 
-  if importlib_bootstrap is not None:
-    _add_finder(importlib_bootstrap.FileFinder, find_wheels_on_path)
+  if importlib_machinery is not None:
+    _add_finder(importlib_machinery.FileFinder, find_wheels_on_path)
 
   __PREVIOUS_FINDER = previous_finder
 
@@ -230,34 +260,37 @@ def unregister_finders():
   pkg_resources.register_finder(zipimport.zipimporter, __PREVIOUS_FINDER)
   _remove_finder(pkgutil.ImpImporter, find_wheels_on_path)
 
-  if importlib_bootstrap is not None:
-    _remove_finder(importlib_bootstrap.FileFinder, find_wheels_on_path)
+  if importlib_machinery is not None:
+    _remove_finder(importlib_machinery.FileFinder, find_wheels_on_path)
 
   __PREVIOUS_FINDER = None
 
 
 def get_script_from_egg(name, dist):
   """Returns location, content of script in distribution or (None, None) if not there."""
-  if name in dist.metadata_listdir('scripts'):
+  if dist.metadata_isdir('scripts') and name in dist.metadata_listdir('scripts'):
     return (
         os.path.join(dist.egg_info, 'scripts', name),
         dist.get_metadata('scripts/%s' % name).replace('\r\n', '\n').replace('\r', '\n'))
   return None, None
 
 
-def safer_name(name):
-  return name.replace('-', '_')
-
-
 def get_script_from_whl(name, dist):
-  # This is true as of at least wheel==0.24.  Might need to take into account the
-  # metadata version bundled with the wheel.
-  wheel_scripts_dir = '%s-%s.data/scripts' % (safer_name(dist.key), dist.version)
-  if dist.resource_isdir(wheel_scripts_dir) and name in dist.resource_listdir(wheel_scripts_dir):
-    script_path = os.path.join(wheel_scripts_dir, name)
-    return (
-        os.path.join(dist.egg_info, script_path),
-        dist.get_resource_string('', script_path).replace(b'\r\n', b'\n').replace(b'\r', b'\n'))
+  # This can get called in different contexts; in some, it looks for files in the
+  # wheel archives being used to produce a pex; in others, it looks for files in the
+  # install wheel directory included in the pex. So we need to look at both locations.
+  datadir_name = WheelMetadata.data_dir(dist.location)
+  wheel_scripts_dirs = ['bin', 'scripts',
+                         os.path.join(datadir_name, "bin"),
+                         os.path.join(datadir_name, "scripts")]
+  for wheel_scripts_dir in wheel_scripts_dirs:
+    if (dist.resource_isdir(wheel_scripts_dir) and
+        name in dist.resource_listdir(wheel_scripts_dir)):
+      # We always install wheel scripts into bin
+      script_path = os.path.join(wheel_scripts_dir, name)
+      return (
+          os.path.join(dist.location, script_path),
+          dist.get_resource_string('', script_path).replace(b'\r\n', b'\n').replace(b'\r', b'\n'))
   return None, None
 
 
@@ -288,17 +321,29 @@ def get_script_from_distributions(name, dists):
 
 
 def get_entry_point_from_console_script(script, dists):
-  # check all distributions for the console_script "script"
-  entries = frozenset(filter(None, (
-      dist.get_entry_map().get('console_scripts', {}).get(script) for dist in dists)))
+  # Check all distributions for the console_script "script". De-dup by dist key to allow for a
+  # duplicate console script IFF the distribution is platform-specific and this is a multi-platform
+  # pex.
+  def get_entrypoint(dist):
+    script_entry = dist.get_entry_map().get('console_scripts', {}).get(script)
+    if script_entry is not None:
+      # Entry points are of the form 'foo = bar', we just want the 'bar' part.
+      return str(script_entry).split('=')[1].strip()
+
+  entries = {}
+  for dist in dists:
+    entry_point = get_entrypoint(dist)
+    if entry_point is not None:
+      entries[dist.key] = (dist, entry_point)
 
-  # if multiple matches, freak out
   if len(entries) > 1:
     raise RuntimeError(
-        'Ambiguous script specification %s matches multiple entry points:%s' % (
-            script, ' '.join(map(str, entries))))
+        'Ambiguous script specification %s matches multiple entry points:\n\t%s' % (
+            script,
+            '\n\t'.join('%r from %r' % (entry_point, dist)
+                        for dist, entry_point in entries.values())))
 
+  dist, entry_point = None, None
   if entries:
-    entry_point = next(iter(entries))
-    # entry points are of the form 'foo = bar', we just want the 'bar' part:
-    return str(entry_point).split('=')[1].strip()
+    dist, entry_point = next(iter(entries.values()))
+  return dist, entry_point
