@@ -4,8 +4,8 @@
 from __future__ import absolute_import, print_function
 
 import os
+import pkgutil
 import sys
-from contextlib import contextmanager
 from distutils import sysconfig
 from site import USER_SITE
 
@@ -21,6 +21,7 @@ from .interpreter import PythonInterpreter
 from .orderedset import OrderedSet
 from .pex_info import PexInfo
 from .tracer import TRACER
+from .util import iter_pth_paths, merge_split, named_temporary_file
 from .variables import ENV
 
 
@@ -31,12 +32,16 @@ class DevNull(object):
   def write(self, *args, **kw):
     pass
 
+  def flush(self):
+    pass
+
 
 class PEX(object):  # noqa: T000
   """PEX, n. A self-contained python environment."""
 
   class Error(Exception): pass
   class NotFound(Error): pass
+  class InvalidEntryPoint(Error): pass
 
   @classmethod
   def clean_environment(cls):
@@ -49,7 +54,7 @@ class PEX(object):  # noqa: T000
     for key in filter_keys:
       del os.environ[key]
 
-  def __init__(self, pex=sys.argv[0], interpreter=None, env=ENV):
+  def __init__(self, pex=sys.argv[0], interpreter=None, env=ENV, verify_entry_point=False):
     self._pex = pex
     self._interpreter = interpreter or PythonInterpreter.get()
     self._pex_info = PexInfo.from_pex(self._pex)
@@ -57,6 +62,8 @@ class PEX(object):  # noqa: T000
     self._vars = env
     self._envs = []
     self._working_set = None
+    if verify_entry_point:
+      self._do_entry_point_verification()
 
   def _activate(self):
     if not self._working_set:
@@ -65,13 +72,18 @@ class PEX(object):  # noqa: T000
       # set up the local .pex environment
       pex_info = self._pex_info.copy()
       pex_info.update(self._pex_info_overrides)
-      self._envs.append(PEXEnvironment(self._pex, pex_info))
-
-      # set up other environments as specified in PEX_PATH
-      for pex_path in filter(None, self._vars.PEX_PATH.split(os.pathsep)):
-        pex_info = PexInfo.from_pex(pex_path)
-        pex_info.update(self._pex_info_overrides)
-        self._envs.append(PEXEnvironment(pex_path, pex_info))
+      pex_info.merge_pex_path(self._vars.PEX_PATH)
+      self._envs.append(PEXEnvironment(self._pex, pex_info, interpreter=self._interpreter))
+      # N.B. by this point, `pex_info.pex_path` will contain a single pex path
+      # merged from pex_path in `PEX-INFO` and `PEX_PATH` set in the environment.
+      # `PEX_PATH` entries written into `PEX-INFO` take precedence over those set
+      # in the environment.
+      if pex_info.pex_path:
+        # set up other environments as specified in pex_path
+        for pex_path in filter(None, pex_info.pex_path.split(os.pathsep)):
+          pex_info = PexInfo.from_pex(pex_path)
+          pex_info.update(self._pex_info_overrides)
+          self._envs.append(PEXEnvironment(pex_path, pex_info, interpreter=self._interpreter))
 
       # activate all of them
       for env in self._envs:
@@ -85,18 +97,38 @@ class PEX(object):  # noqa: T000
   @classmethod
   def _extras_paths(cls):
     standard_lib = sysconfig.get_python_lib(standard_lib=True)
+
     try:
       makefile = sysconfig.parse_makefile(sysconfig.get_makefile_filename())
     except (AttributeError, IOError):
       # This is not available by default in PyPy's distutils.sysconfig or it simply is
       # no longer available on the system (IOError ENOENT)
       makefile = {}
+
     extras_paths = filter(None, makefile.get('EXTRASPATH', '').split(':'))
     for path in extras_paths:
       yield os.path.join(standard_lib, path)
 
-  @classmethod
-  def _get_site_packages(cls):
+    # Handle .pth injected paths as extras.
+    sitedirs = cls._get_site_packages()
+    for pth_path in cls._scan_pth_files(sitedirs):
+      TRACER.log('Found .pth file: %s' % pth_path, V=3)
+      for extras_path in iter_pth_paths(pth_path):
+        yield extras_path
+
+  @staticmethod
+  def _scan_pth_files(dir_paths):
+    """Given an iterable of directory paths, yield paths to all .pth files within."""
+    for dir_path in dir_paths:
+      if not os.path.exists(dir_path):
+        continue
+
+      pth_filenames = (f for f in os.listdir(dir_path) if f.endswith('.pth'))
+      for pth_filename in pth_filenames:
+        yield os.path.join(dir_path, pth_filename)
+
+  @staticmethod
+  def _get_site_packages():
     try:
       from site import getsitepackages
       return set(getsitepackages())
@@ -141,6 +173,11 @@ class PEX(object):  # noqa: T000
         new_modules[module_name] = module
         continue
 
+      # Unexpected objects, e.g. namespace packages, should just be dropped:
+      if not isinstance(module.__path__, list):
+        TRACER.log('Dropping %s' % (module_name,), V=3)
+        continue
+
       # Pop off site-impacting __path__ elements in-place.
       for k in reversed(range(len(module.__path__))):
         if cls._tainted_path(module.__path__[k], site_libs):
@@ -172,7 +209,7 @@ class PEX(object):  # noqa: T000
 
     user_site_distributions.update(all_distribution_paths(USER_SITE))
 
-    if not inherit_path:
+    if inherit_path == 'false':
       scrub_paths = site_distributions | user_site_distributions
       for path in user_site_distributions:
         TRACER.log('Scrubbing from user site: %s' % path)
@@ -212,30 +249,17 @@ class PEX(object):  # noqa: T000
     return sys_path, sys_path_importer_cache, sys_modules
 
   @classmethod
-  @contextmanager
   def patch_pkg_resources(cls, working_set):
     """Patch pkg_resources given a new working set."""
-    def patch(working_set):
-      pkg_resources.working_set = working_set
-      pkg_resources.require = working_set.require
-      pkg_resources.iter_entry_points = working_set.iter_entry_points
-      pkg_resources.run_script = pkg_resources.run_main = working_set.run_script
-      pkg_resources.add_activation_listener = working_set.subscribe
+    pkg_resources.working_set = working_set
+    pkg_resources.require = working_set.require
+    pkg_resources.iter_entry_points = working_set.iter_entry_points
+    pkg_resources.run_script = pkg_resources.run_main = working_set.run_script
+    pkg_resources.add_activation_listener = working_set.subscribe
 
-    old_working_set = pkg_resources.working_set
-    patch(working_set)
-    try:
-      yield
-    finally:
-      patch(old_working_set)
-
-  # Thar be dragons -- when this contextmanager exits, the interpreter is
-  # potentially in a wonky state since the patches here (minimum_sys_modules
-  # for example) actually mutate global state.  This should not be
-  # considered a reversible operation despite being a contextmanager.
-  @classmethod
-  @contextmanager
-  def patch_sys(cls, inherit_path):
+  # Thar be dragons -- when this function exits, the interpreter is potentially in a wonky state
+  # since the patches here (minimum_sys_modules for example) actually mutate global state.
+  def patch_sys(self, inherit_path):
     """Patch sys with all site scrubbed."""
     def patch_dict(old_value, new_value):
       old_value.clear()
@@ -246,17 +270,15 @@ class PEX(object):  # noqa: T000
       patch_dict(sys.path_importer_cache, path_importer_cache)
       patch_dict(sys.modules, modules)
 
-    old_sys_path, old_sys_path_importer_cache, old_sys_modules = (
-        sys.path[:], sys.path_importer_cache.copy(), sys.modules.copy())
-    new_sys_path, new_sys_path_importer_cache, new_sys_modules = cls.minimum_sys(inherit_path)
+    new_sys_path, new_sys_path_importer_cache, new_sys_modules = self.minimum_sys(inherit_path)
+
+    new_sys_path.extend(merge_split(self._pex_info.pex_path, self._vars.PEX_PATH))
 
     patch_all(new_sys_path, new_sys_path_importer_cache, new_sys_modules)
-    yield
 
   def _wrap_coverage(self, runner, *args):
     if not self._vars.PEX_COVERAGE and self._vars.PEX_COVERAGE_FILENAME is None:
-      runner(*args)
-      return
+      return runner(*args)
 
     try:
       import coverage
@@ -273,7 +295,7 @@ class PEX(object):  # noqa: T000
     cov.start()
 
     try:
-      runner(*args)
+      return runner(*args)
     finally:
       TRACER.log('Stopping coverage')
       cov.stop()
@@ -287,8 +309,7 @@ class PEX(object):  # noqa: T000
 
   def _wrap_profiling(self, runner, *args):
     if not self._vars.PEX_PROFILE and self._vars.PEX_PROFILE_FILENAME is None:
-      runner(*args)
-      return
+      return runner(*args)
 
     pex_profile_filename = self._vars.PEX_PROFILE_FILENAME
     pex_profile_sort = self._vars.PEX_PROFILE_SORT
@@ -307,6 +328,10 @@ class PEX(object):  # noqa: T000
       else:
         profiler.print_stats(sort=pex_profile_sort)
 
+  def path(self):
+    """Return the path this PEX was built at."""
+    return self._pex
+
   def execute(self):
     """Execute the PEX.
 
@@ -315,15 +340,15 @@ class PEX(object):  # noqa: T000
     """
     teardown_verbosity = self._vars.PEX_TEARDOWN_VERBOSE
     try:
-      pex_inherit_path = self._vars.PEX_INHERIT_PATH or self._pex_info.inherit_path
-      with self.patch_sys(pex_inherit_path):
-        working_set = self._activate()
-        TRACER.log('PYTHONPATH contains:')
-        for element in sys.path:
-          TRACER.log('  %c %s' % (' ' if os.path.exists(element) else '*', element))
-        TRACER.log('  * - paths that do not exist or will be imported via zipimport')
-        with self.patch_pkg_resources(working_set):
-          self._wrap_coverage(self._wrap_profiling, self._execute)
+      pex_inherit_path = self._vars.PEX_INHERIT_PATH
+      if pex_inherit_path == "false":
+        pex_inherit_path = self._pex_info.inherit_path
+      self.patch_sys(pex_inherit_path)
+      working_set = self._activate()
+      self.patch_pkg_resources(working_set)
+      exit_code = self._wrap_coverage(self._wrap_profiling, self._execute)
+      if exit_code:
+        sys.exit(exit_code)
     except Exception:
       # Allow the current sys.excepthook to handle this app exception before we tear things down in
       # finally, then reraise so that the exit status is reflected correctly.
@@ -371,34 +396,88 @@ class PEX(object):  # noqa: T000
       TRACER.log('No entry point specified, dropping into interpreter')
       return self.execute_interpreter()
 
+  @classmethod
+  def demote_bootstrap(cls):
+    TRACER.log('Bootstrap complete, performing final sys.path modifications...')
+
+    bootstrap_path = __file__
+    module_import_path = __name__.split('.')
+    root_package = module_import_path[0]
+
+    # For example, our __file__ might be requests.pex/.bootstrap/_pex/pex.pyc and our import path
+    # _pex.pex; so we walk back through all the module components of our import path to find the
+    # base sys.path entry where we were found (requests.pex/.bootstrap in this example).
+    for _ in module_import_path:
+      bootstrap_path = os.path.dirname(bootstrap_path)
+    bootstrap_path_index = sys.path.index(bootstrap_path)
+
+    # Move the third party resources pex uses to the end of sys.path for the duration of the run to
+    # allow conflicting versions supplied by user dependencies to win during the course of the
+    # execution of user code.
+    for _, mod, _ in pkgutil.iter_modules([bootstrap_path]):
+      if mod != root_package:  # We let _pex stay imported
+        TRACER.log('Un-importing third party bootstrap dependency %s from %s'
+                   % (mod, bootstrap_path))
+        sys.modules.pop(mod)
+        submod_prefix = mod + '.'
+        for submod in [m for m in sys.modules.keys() if m.startswith(submod_prefix)]:
+          sys.modules.pop(submod)
+    sys.path.pop(bootstrap_path_index)
+    sys.path.append(bootstrap_path)
+
+    TRACER.log('PYTHONPATH contains:')
+    for element in sys.path:
+      TRACER.log('  %c %s' % (' ' if os.path.exists(element) else '*', element))
+    TRACER.log('  * - paths that do not exist or will be imported via zipimport')
+
   def execute_interpreter(self):
-    if sys.argv[1:]:
-      try:
-        with open(sys.argv[1]) as fp:
-          name, content = sys.argv[1], fp.read()
-      except IOError as e:
-        die("Could not open %s in the environment [%s]: %s" % (sys.argv[1], sys.argv[0], e))
-      sys.argv = sys.argv[1:]
-      self.execute_content(name, content)
+    args = sys.argv[1:]
+    if args:
+      # NB: We take care here to setup sys.argv to match how CPython does it for each case.
+      arg = args[0]
+      if arg == '-c':
+        content = args[1]
+        sys.argv = ['-c'] + args[2:]
+        self.execute_content('-c <cmd>', content, argv0='-c')
+      elif arg == '-m':
+        module = args[1]
+        sys.argv = args[1:]
+        self.execute_module(module)
+      else:
+        try:
+          if arg == '-':
+            content = sys.stdin.read()
+          else:
+            with open(arg) as fp:
+              content = fp.read()
+        except IOError as e:
+          die("Could not open %s in the environment [%s]: %s" % (arg, sys.argv[0], e))
+        sys.argv = args
+        self.execute_content(arg, content)
     else:
+      self.demote_bootstrap()
+
       import code
       code.interact()
 
   def execute_script(self, script_name):
     dists = list(self._activate())
 
-    entry_point = get_entry_point_from_console_script(script_name, dists)
+    dist, entry_point = get_entry_point_from_console_script(script_name, dists)
     if entry_point:
+      TRACER.log('Found console_script %r in %r' % (entry_point, dist))
       sys.exit(self.execute_entry(entry_point))
 
     dist, script_path, script_content = get_script_from_distributions(script_name, dists)
     if not dist:
-      raise self.NotFound('Could not find script %s in pex!' % script_name)
-    TRACER.log('Found script %s in %s' % (script_name, dist))
+      raise self.NotFound('Could not find script %r in pex!' % script_name)
+    TRACER.log('Found script %r in %r' % (script_name, dist))
     return self.execute_content(script_path, script_content, argv0=script_name)
 
   @classmethod
   def execute_content(cls, name, content, argv0=None):
+    cls.demote_bootstrap()
+
     argv0 = argv0 or name
     try:
       ast = compile(content, name, 'exec', flags=0, dont_inherit=1)
@@ -426,13 +505,17 @@ class PEX(object):  # noqa: T000
     runner = cls.execute_pkg_resources if ':' in entry_point else cls.execute_module
     return runner(entry_point)
 
-  @staticmethod
-  def execute_module(module_name):
+  @classmethod
+  def execute_module(cls, module_name):
+    cls.demote_bootstrap()
+
     import runpy
     runpy.run_module(module_name, run_name='__main__')
 
-  @staticmethod
-  def execute_pkg_resources(spec):
+  @classmethod
+  def execute_pkg_resources(cls, spec):
+    cls.demote_bootstrap()
+
     entry = EntryPoint.parse("run = {0}".format(spec))
 
     # See https://pythonhosted.org/setuptools/history.html#id25 for rationale here.
@@ -479,3 +562,32 @@ class PEX(object):  # noqa: T000
                                     stderr=kwargs.pop('stderr', None),
                                     **kwargs)
     return process.wait() if blocking else process
+
+  def _do_entry_point_verification(self):
+
+    entry_point = self._pex_info.entry_point
+    ep_split = entry_point.split(':')
+
+    # a.b.c:m ->
+    # ep_module = 'a.b.c'
+    # ep_method = 'm'
+
+    # Only module is specified
+    if len(ep_split) == 1:
+      ep_module = ep_split[0]
+      import_statement = 'import {}'.format(ep_module)
+    elif len(ep_split) == 2:
+      ep_module = ep_split[0]
+      ep_method = ep_split[1]
+      import_statement = 'from {} import {}'.format(ep_module, ep_method)
+    else:
+      raise self.InvalidEntryPoint("Failed to parse: `{}`".format(entry_point))
+
+    with named_temporary_file() as fp:
+      fp.write(import_statement.encode('utf-8'))
+      fp.close()
+      retcode = self.run([fp.name], env={'PEX_INTERPRETER': '1'})
+      if retcode != 0:
+        raise self.InvalidEntryPoint('Invalid entry point: `{}`\n'
+                                'Entry point verification failed: `{}`'
+                                .format(entry_point, import_statement))
