@@ -9,11 +9,13 @@ import shutil
 import tempfile
 import uuid
 from hashlib import sha1
+from site import makepath
 from threading import Lock
 
 from pkg_resources import find_distributions, resource_isdir, resource_listdir, resource_string
 
 from .common import rename_if_empty, safe_mkdir, safe_mkdtemp, safe_open
+from .compatibility import exec_function
 from .finders import register_finders
 
 
@@ -89,9 +91,9 @@ class DistributionHelper(object):
     # Monkeypatch pkg_resources finders should it not already be so.
     register_finders()
     if name is None:
-      distributions = list(find_distributions(path))
+      distributions = set(find_distributions(path))
       if len(distributions) == 1:
-        return distributions[0]
+        return distributions.pop()
     else:
       for dist in find_distributions(path):
         if dist.project_name == name:
@@ -213,3 +215,44 @@ def named_temporary_file(*args, **kwargs):
       yield fp
   finally:
     os.remove(fp.name)
+
+
+def iter_pth_paths(filename):
+  """Given a .pth file, extract and yield all inner paths without honoring imports. This shadows
+  python's site.py behavior, which is invoked at interpreter startup."""
+  try:
+    f = open(filename, 'rU')  # noqa
+  except IOError:
+    return
+
+  dirname = os.path.dirname(filename)
+  known_paths = set()
+
+  with f:
+    for line in f:
+      line = line.rstrip()
+      if not line or line.startswith('#'):
+        continue
+      elif line.startswith(('import ', 'import\t')):
+        try:
+          exec_function(line)
+          continue
+        except Exception:
+          # Defer error handling to the higher level site.py logic invoked at startup.
+          return
+      else:
+        extras_dir, extras_dir_case_insensitive = makepath(dirname, line)
+        if extras_dir_case_insensitive not in known_paths and os.path.exists(extras_dir):
+          yield extras_dir
+          known_paths.add(extras_dir_case_insensitive)
+
+
+def merge_split(*paths):
+  """Merge paths into a single path delimited by colons and split on colons to return
+  a list of paths.
+
+  :param paths: a variable length list of path strings
+  :return: a list of paths from the merged path list split by colons
+  """
+  filtered_paths = filter(None, paths)
+  return [p for p in ':'.join(filtered_paths).split(':') if p]
