@@ -26,13 +26,23 @@ def test_empty_iteration():
   assert kwargs == {'follow_links': False}
 
 
-def test_iteration_with_return():
-  pex_url = 'https://pypi.python.org/packages/source/p/pex/pex-0.8.6.tar.gz'
-  crawler_mock = mock.create_autospec(Crawler, spec_set=True)
-  crawler_mock.crawl.return_value = [pex_url]
-  iterator = Iterator(crawler=crawler_mock, follow_links=True)
+def assert_iteration(all_versions, *expected_versions, **iterator_kwargs):
+  def package_url(version):
+    return 'https://pypi.org/packages/source/p/pex/pex-%s.tar.gz' % version
 
-  assert list(iterator.iter(Requirement.parse('pex'))) == [SourcePackage(pex_url)]
+  urls = [package_url(v) for v in all_versions]
+  crawler_mock = mock.create_autospec(Crawler, spec_set=True)
+  crawler_mock.crawl.return_value = urls
+  iterator = Iterator(crawler=crawler_mock, follow_links=True, **iterator_kwargs)
+
+  assert list(iterator.iter(Requirement.parse('pex'))) == [SourcePackage(package_url(v))
+                                                           for v in expected_versions]
   assert len(crawler_mock.crawl.mock_calls) == 1
   _, _, kwargs = crawler_mock.crawl.mock_calls[0]
   assert kwargs == {'follow_links': True}
+
+
+def test_iteration_with_return():
+  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6')  # Iterator should exclude prereleases by default.
+  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6', allow_prereleases=False)
+  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6', '0.8.6b1', allow_prereleases=True)
